(defpackage prop-logic
  (:documentation "Package for representing formulas in propositional logic.

Formulas are representing using the [formula] type.  Elementary propositions are represented by symbols or lists (denoting predicates), and t and nil represent the logical constants TRUE and FALSE.  More complex formulae are built out of these using conjoin, disjoin, or negate.  A propositional state (aka possible world aka model), using the closed world assumption, is represented by a list of propositions that are true in that state.

Basic creation and access
-------------------------
conjoin
conjoin-list
disjoin
disjoin-list
negate
conjuncts
disjuncts
negatee
literal-prop
standardize-prop
standardize-literal
prop-symbol
prop-args

Types
-----
is-proposition
is-state
is-formula
[formula]
is-literal
is-dnf-clause
is-dnf-formula
is-negation
is-conjunction
is-disjunction

Comparisons 
-----------
prop<
prop-equal
literal<
literal-equal

General logic
-------------
holds
implies
consistent

Substitution 
------------
bind

Parametrized sets of propositions
---------------------------------
make-prop-set

DNF
---
standardize-dnf
sort-dnf
simplify-dnf
inter-clause-simplify
dnf-or
dnf-and
dnf-not
dnf-clauses
make-state-dnf

DNF sets
--------
<dnf-set>
make-dnf-set
formula
props
size-must-exceed

Debugging
---------
pprint-dnf
with-dnf-pprint

   ")
  

  (:export
   conjoin
   conjoin-list
   disjoin
   disjoin-list
   negate
   conjuncts
   disjuncts
   negatee
   literal-prop
   standardize-literal
   standardize-prop
   prop-symbol
   prop-args
   
   is-proposition
   is-state
   is-formula
   [formula]
   is-literal
   is-negation
   is-conjunction
   is-disjunction

   prop<
   prop-equal
   literal<
   literal-equal
   
   holds
   implies
   consistent
   
   bind
   
   make-prop-set

   
   standardize-dnf
   sort-dnf
   simplify-dnf
   inter-clause-simplify
   dnf-or
   dnf-and
   dnf-not
   is-dnf-clause
   is-dnf-formula
   dnf-clauses
   make-state-dnf
   
   make-dnf-set
   <dnf-set>
   formula
   props
   size-must-exceed
   
   pprint-dnf
   with-dnf-pprint
   )
  (:use
   cl
   utils
   inst-vars
   prod-set
   set))
   

(in-package prop-logic)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; various basic types of formulae
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(declaim (inline is-constant is-proposition is-compound-formula
		 is-conjunction is-disjunction is-negation is-literal))
		 

;; TODO these should be changed to deftypes

(defun is-constant (x)
  "is-constant X.  Is X one of t or nil?"
  (member x '(t nil)))

(defun is-proposition (x)
  "is-proposition X.  X is a proposition if it is a symbol that is not t or nil, or if it is a list whose first element is not 'not, 'or, or 'and."
  (or (and (symbolp x) (not (is-constant x)))
      (and (consp x) (not (member (first x) '(not and or))))))

(defun is-compound-formula (x)
  "is-compound-formula X.  Is X an and, not, or or (may not recursively check all the arguments for well-formedness)?"
  (to-boolean 
   (and (consp x)
	(member (car x) '(or not and)))))

(defun is-conjunction (x)
  "is-conjunction X.  Is X an and (does not recursively check arguments for wellformedness)?"
  (and (consp x) (eq (car x) 'and)))

(defun is-disjunction (x)
  "Is X an or (does not recursively check arguments for wellformedness)?"
  (and (consp x) (eq (car x) 'or)))

(defun is-negation (x)
  "Is X a not (does not recursively check arguments for wellformedness)?"
  (and (listp x) (eq (car x) 'not) (cdr x) (not (cddr x))))

(defun compound-formula-type (x)
  "compound-formula X.  For a compound-formula, return the type ('or, 'not, or 'and) (answer undefined if not a compound formula)."
  (let ((f (car x)))
    (if (member f '(not and or))
	f
      'not-compound-formula)))

(defun is-literal (x)
  "is-literal X.  Is X a proposition or negated proposition?"
  (or (is-proposition x)
      (and (is-negation x)
	   (is-proposition (negatee x)))))






;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; basic creation and access
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(declaim (inline conjoin conjoin-list disjoin disjoin-list negate negatee
		 conjuncts disjuncts prop-symbol prop-args literal-prop
		 standardize-literal standardize-prop))

(defun conjoin (&rest args)
  "conjoin &rest FORMULAE.  Return a formula representing the logical and of the FORMULAE."
  (cons 'and args))

(defun conjoin-list (l)
  "conjoin-list L.  Return a formula representing the logical and of the formulae in list L."
  (cons 'and l))

(defun disjoin (&rest args)
  "disjoin &rest FORMULAE.  Return a formula representing the logical or of the FORMULAE."
  (cons 'or args))

(defun disjoin-list (l)
  "disjoin-list L.  Return a formula representing the logical or of the formulae in list L."
  (cons 'or l))

(defun negate (f)
  "negate FORMULA.  Return the negation of FORMULA, and also do simplification to remove double negatives."
  (if (is-negation f)
      (negatee f)
    (list 'not f)))

(defun negatee (x)
  "negatee X.  For a negation, return the thing that's being negated (answer undefined if not a negation)."
  (second x))

(defun conjuncts (x)
  "conjuncts X.  If X is a conjunction, return a list of the conjuncts, otherwise just return a list containing X."
  (if (is-literal x)
      (list x)
    (rest x)))

(defun disjuncts (x)
  "disjuncts X.  For a disjunction (including a single proposition), return list of formulae being disjoined.  Not a fresh list, so copy it before modifying.  Answer undefined if not a disjunction."
  (if (is-proposition x)
      (list x)
    (rest x)))

(defun prop-symbol (p)
  "prop-symbol P.  If P is a symbol return it.  Otherwise, assume it's a list and return its first element."
  (etypecase p
    (symbol p)
    (cons (first p))))

(defun prop-args (p)
  "prop-args P.  If P is a symbol, nil, otherwise, assume its a list and return (rest p)."
  (etypecase p
    (symbol p)
    (cons (rest p))))

(defun literal-prop (l)
  "literal-prop L.  Get the proposition used in L.  Answer undefined if not a literal."
  (if (is-proposition l) l
    (negatee l)))

(defun standardize-prop (prop)
  (if (or (is-constant prop) (listp prop))
      prop
    (list prop)))

(defun standardize-literal (l)
  "standardize-literal L.  If the proposition in L is just a symbol, enclose it in a list."

  (if (is-negation l)
      (negate (standardize-prop (literal-prop l)))
    (standardize-prop l)))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; general formulae
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun is-formula (x)
  "is-formula X

Returns true iff X is one of
1) t
2) nil
3) A symbol, or list whose first element is not 'not, 'and, or 'or and whose remaining elements are symbols or numbers
4) A list '(not F) where F is a formula
5) A list '(S . F) where F is a list of formulae and S is either 'and or 'or"
  (or
   (is-constant x)
   (and
    (is-proposition x)
    (or (symbolp x)
	(and (consp x)
	     (loop for j on (cdr x)
		 when (null j) return t
		 unless (consp j) return nil
		 unless (or (symbolp (first j)) (numberp (first j))) return nil
		 finally (return t)))))
   (cond
    ((is-negation x) (is-formula (negatee x)))
    ((is-conjunction x) (every #'is-formula (conjuncts x)))
    ((is-disjunction x) (every #'is-formula (disjuncts x))))))

(deftype [formula] ()
  "Type for propositional formulae.  Well-formed formulae are determined using the predicate is-formula.  See its documentation for more."
  `(satisfies is-formula))


(defun propositions (formula &optional (result-type 'list))
  "propositions FORMULA &optional (RESULT-TYPE 'list).  Return the set of elementary propositions used in this FORMULA.  RESULT-TYPE can only be 'list for now."
  (assert (eql result-type 'list))
  (let ((l nil))
    (labels 
	((helper (f)
	   (cond
	    ((is-literal f)
	     (adjoin (literal-prop f) l))
	    ((is-compound-formula f)
	     (ecase (compound-formula-type f)
	       (not (helper (negatee f)))
	       (and (map nil #'helper (conjuncts f)))
	       (or (map nil #'helper (disjuncts f))))))
	   (values)))
      (helper formula))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; comparison
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(declaim (inline prop-equal literal-equal))

(defun prop-equal (f1 f2)
  "Check if propositions F1 and F2 are equal.  If they are lists, then #'equal is used.  If F_i is a symbol, it is interpreted as (list Fi)."
  (if (symbolp f1)
      (if (symbolp f2)
	  (eq f2 f1)
	(unless (length-exceeds f2 1)
	  (eq (first f2) f1)))
    (if (symbolp f2)
	(unless (length-exceeds f1 1)
	  (eq (first f1) f2))
      (equal f1 f2))))

(defun literal-equal (l1 l2)
  "Check if literals L1 and L2 are equal.  Use prop-equal."
  (if (is-negation l1)
      (and (is-negation l2) (prop-equal (negatee l1) (negatee l2)))
    (and (not (is-negation l2)) (prop-equal l1 l2))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; states
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun is-state (s)
  "A state aka model aka interpretation, is represented as the set of propositions that are true in it.  Propositions must be represented as lists, e.g. the shorthand of writing 'foo for '(foo) cannot be used."
  (each s #'is-proposition))
  


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; satisfaction
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defgeneric holds (state formula)
  (:documentation "holds STATE FORMULA.  Does STATE make FORMULA true?  STATE is the set of propositions that are true.")
  (:method (state (formula (eql t)))
	   (declare (ignore state))
	   t)
  (:method (state (formula (eql nil)))
	   (declare (ignore state))
	   nil)
  (:method (state (formula symbol))
	   (member? (list formula) state))
  (:method (state (formula list))
	   (case (first formula)
	     (and (every #'(lambda (f) (holds state f)) (rest formula)))
	     (or (some (lambda (f) (holds state f)) (rest formula)))
	     (not (not (holds state (second formula))))
	     (otherwise (member? formula state)))))

(defun implies (f1 f2)
  "implies F1 F2.  Logical implication.  For now only works for DNF formulae."
  (assert (and (is-dnf-formula f1) (is-dnf-formula f2)) ()
    "implies currently only works for DNF formulae.")
  (dnf-implies f1 f2))

(defun consistent (f1 f2)
  "consistent F1 F2.  True iff F1 does not imply not(F2).  Only works for DNF formulae."
  ;; this is probably not the most efficient way to do it
  (assert (and (is-dnf-formula f1) (is-dnf-formula f2)) ()
    "consistent currently only works for DNF formulae.")
  (not (dnf-implies f1 (dnf-not f2))))

  

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; substitution
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun bind (formula bindings)
  "bind FORMULA BINDINGS.  FORMULA is a formula, and BINDINGS is an association list mapping symbols to other symbols or numbers.  Returns a new formula which is identical to FORMULA, except that for every proposition of the form (NAME ARG1 ... ARGn) occurring in FORMULA, for which some of the ARGi are keys in BINDINGS, those arguments are replaced by their corresponding value in BINDINGS.  The substitutions are done in parallel.  The newly returned formula is fresh; in particular, it shares no structure with FORMULA.  "
  (cond
   ((is-constant formula) formula)
   ((is-compound-formula formula) 
    (cons (first formula)
	  (mapcar #'(lambda (f) (bind f bindings)) (rest formula))))
   ((is-proposition formula)
    (if (symbolp formula)
	formula
      (cons (first formula)
	    (mapcar #'(lambda (arg)
			(mvbind (val defined?)
			    (mapping:evaluate-mv bindings arg)
			  (if defined? val arg)))
		    (rest formula)))))
   (t (assert nil nil "Unrecognized formula type for ~a" formula))))
			    
      
  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; sets of propositions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <prop-set> (<prod-set>)
  ((prop :initarg :prop-symbol :accessor prop)
   (domains :initarg :domains :accessor domains)))

(defun make-prop-set (prop-symbol &rest domains)
  "make-prop-set PROP-SYMBOL DOMAIN_1 ... DOMAIN_M

Return a compact representation of the set of propositions of the form (PROP-SYMBOL X_1 ... X_M) where each X_i belongs to the corresponding DOMAIN_i."
  
  (if domains
  
      (let* ((n (length domains))
	     (acc
	      (make-inst-var-accessors
	       :creator 
	       #'(lambda () 
		   (cons prop-symbol 
			 (make-list n :initial-element 'uninstantiated)))
	       :readers
	       (mapset 'vector #'(lambda (i) #'(lambda (l) (nth (1+ i) l))) n)
	       :writers
	       (mapset 'vector #'(lambda (i) #'(lambda (l v) (setf (nth (1+ i) l) v))) n)
	       :invalid-inst-detector
	       #'(lambda (inst)
		   (if (listp inst)
		       (handler-case
			   (if (eq (first inst) prop-symbol)
			       (if (every #'(lambda (x s) (member? x s)) (rest inst) domains)
				   nil
				 'invalid-component)
			     'incorrect-prop-symbol)
			 (simple-error (c)
			   (if (is-true-list inst)
			       (error c)
			     'not-true-list)))
		     'not-list))
	       :var-names n
	       :var-num #'identity)))
	(make-instance '<prop-set> :sets domains :inst-acc acc :prop-symbol prop-symbol :domains domains))
    (list (list prop-symbol))))

(defun pprint-prop-set (str s)
  (pprint-logical-block (str (domains s) :suffix "]")
    (format str "[Prop set with prop symbol ~a and components " (prop s))
    (loop
      (format str "~W" (pprint-pop))
      (pprint-exit-if-list-exhausted)
      (format str ", "))))

(set-pprint-dispatch '<prop-set> #'pprint-prop-set 2)