(in-package lookahead)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; basic type def and accessors
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <hierarchy> ()
  ((planning-problem :initarg :planning-problem :accessor planning-problem)
   (high-level-actions :accessor high-level-actions))
  (:documentation "Class <hierarchy>

Initargs
:planning-problem - of type <planning-problem>
"))


(defgeneric refinements (h a)
  (:documentation "Return the set of legal refinements of high-level action A."))

(defun action-type (a h)
  "action-type ACTION-NAME HIERARCHY.  Returns one of 'high-level, 'primitive, or 'unknown-action."
  (cond
   ((member? a (primitive-actions h)) 'primitive)
   ((member? a (high-level-actions h)) 'high-level)
   (t 'unknown-action)))

(defun primitive-actions (h)
  (all-actions (planning-problem h)))

(defun is-primitive-sequence (a h)
  "is-primitive-sequence ACTION-SEQUENCE HIERARCHY.  Are all the actions primitive?"
  (every #'(lambda (x) (eq (action-type x h) 'primitive)) a))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; getting immediate refinements of a sequence
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun get-immediate-refinements (h actions)
  "get-immediate-refinements HIERARCHY ACTION-SEQ
Get all refinements of ACTION-SEQ obtained by selecting one of the abstract actions in it and choosing one of its refinements."
  ;; TODO - is this ever used?
  (disjoint-union-list
   (mapset 
    'list
    #'(lambda (i)
	(let ((a (elt actions i)))
	  (when (eq (action-type a h) 'high-level)
	    (make-image-set 
	     (refinements h a) 
	     #'(lambda (ref) 
		 (concatenate 'vector (subseq actions 0 i) ref (subseq actions (1+ i))))
	     #'(lambda (seq)
		 (let ((l (1+ (- (length seq) (length actions)))))
		   (subseq seq i (+ i l))))))))
    (length actions))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; called by the old version
;; (sound-complete-forward-search
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun immediate-applicable-refinements (h actions hsets &key ref-gen)
  "immediate-applicable-refinements HIERARCHY ACTION-SEQ HSETS &key REF-GEN
Return all refinements obtained by picking  some abstract action in the sequence and choosing one of its applicable refinements, where applicability is determined by calling applicable-refinements on that action for the given sound and complete sets.

Orders refinement randomly.

REFGEN can either be 'product or 'first, corresponding to different ways of generating the set of refinements.
"
  (random-permutation
   (funcall
    (ecase ref-gen
      (product #'cross-product-refinements)
      (first #'first-hla-refinements))
    h actions hsets)))

(defun cross-product-refinements (h actions hsets)
  (make-image-set
   (apply #'direct-product 'list
	  (mapset 'list #'(lambda (i) 
			    (let* ((a (elt actions i))
				   (just-a (list (list a))))
			      (if (eq (action-type a h) 'high-level)
				  (or (applicable-refinements h a (aref hsets i)) just-a)
				just-a)))
		  (length actions)))
   #'(lambda (refs) (apply #'concatenate 'list refs))))

(defun first-hla-refinements (h actions hsets)
  (mvbind (i key a) (argmin actions :key #'(lambda (a) (ref-level h a)))
    (declare (ignore key))
    (when (eq (action-type a h) 'high-level)
      (awhen (applicable-refinements h a (aref hsets i))
	     (make-image-set 
	      it #'(lambda (ref) (concatenate 'list (subseq actions 0 i) ref (subseq actions (1+ i))))
	      #'(lambda (refined-actions) (subseq refined-actions i (+ i 1 (- (length refined-actions) (length actions))))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Called by hierarchical-forward-search
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter *random-refinement-ordering* t
  "Variable *random-refinement-ordering*.  Defaults to t.  If true, then the set of refinements of a plan is (uniformly) randomly permuted each time.")

(defun immediate-applicable-refinements-and-depths (h actions depths hsets)
  "immediate-applicable-refinements-and-depths HIERARCHY ACTION-SEQ HSETS"
  (mvbind (i key a) (argmin actions :key #'(lambda (a) (ref-level h a)))
    (declare (ignore key))
    (let ((depth (elt depths i)))
      (when (eq (action-type a h) 'high-level)
	(ndlet ((ref (funcall (if *random-refinement-ordering* #'random-permutation #'identity)
			      (applicable-refinements h a (aref hsets i)))))
	  (let ((new-depths (make-array (length ref) :element-type 'fixnum :initial-element (1+ depth))))
	    (cons
	     (concatenate 'list (subseq actions 0 i) ref (subseq actions (1+ i)))
	     (concatenate 'list (subseq depths 0 i) new-depths (subseq depths (1+ i))))))))))




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; other generic functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defgeneric applicable-refinements (h a hset)
  (:documentation "applicable-refinements HIERARCHY ACTION HSET.  Return the set of refinements of ACTION that are 'applicable' given this hset.  As with applicable-top-level-actions, the actual meaning may vary.")
  (:method ((h <hierarchy>) a hset)
	   (declare (ignore hset))
	   (refinements h a)))
  

(defgeneric applicable-top-level-actions (h hset)
  (:documentation "applicable-top-level-actions HIERARCHY HSET.  Return the set of 'applicable' top level actions given this hset.  The meaning of 'applicable' may vary.  For example, given a precondition function for actions, it may return the set of actions whose preconditions are satisfied at some state in the set.  The default method returns all the high-level and primitive actions.")
  (:method ((h <hierarchy>) hset)
	   (declare (ignore hset))
	   (disjoint-union (high-level-actions h) (primitive-actions h))))


(defgeneric ref-level (h a)
  (:documentation "ref-level H A.  What is the refinement level of this action?  Used for controlling when things get refined.")
  (:method ((h <hierarchy>) a) (declare (ignore a)) 0))