(in-package lookahead)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; class def
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <planning-problem> (<fully-observable-env>)
  ()
  (:documentation "Abstract class <planning-problem> (<env>)

Subclasses: <prop-domain>

A planning problem is viewed as an environment where the initial state and transitions are deterministic, a reward of -1 is given per step, and the episode terminates when a goal state is reached.

Subclasses must implement
- goal
- init-state
- primitive-action-description
- avail-actions
- init-state-set
- same-state

")
  (:default-initargs :initialize-state nil))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; constructor
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod initialize-instance :around ((d <planning-problem>) &rest args)
  (declare (ignore args))
  (call-next-method)
  (create-env:set-state (init-state d) d))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; basic operations
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun action-result (e s a)
  "action-result E S A.  Resulting state after doing primitive action A in state S of env E."
  (succ-state s (primitive-action-description e a)))

(defgeneric goal (prob)
  (:documentation "goal PROB.  Return goal set of PROB."))

(defun goal? (prob s)
  (values
   (member? s (goal prob))))

(defgeneric init-state (e)
  (:documentation "init-state ENV."))

(defun init-state-set (e)
  "init-state-set PLANNING-DOMAIN.  Return a representation of the singleton set consisting of the initial state of this domain."
  (make-state-set e (init-state e)))

(defgeneric make-state-set (e s)
  (:documentation "make-state-set DOMAIN STATE.  Return a representation of the singleton set consisting of this state."))
  

(defun action-seq-result (e s actions)
  "action-seq-result E S A.  Return resulting state if we do (primitive) actions in A starting in state S."
  (map nil (lambda (a) (setf s (action-result e s a))) actions)
  s)

(defgeneric primitive-action-description (e a)
  (:documentation "primitive-action-description PLANNING-DOMAIN ACTION.  Get description of this action."))

(defun succeeds? (e actions &optional (s (init-state e)))
  "succeeds? E PRIMITIVE-ACTION-SEQUENCE STATE.  Does doing this action sequence in the initial state result in a state that satisfies the goal?  STATE defaults to the initial state."
  (goal? e (action-seq-result e s actions)))

(defgeneric all-actions (e)
  (:documentation "all-actions PLANNING-PROBLEM.  Get set of all possible primitive-actions in PLANNING-PROBLEM."))

(defgeneric same-state (s1 s2)
  (:documentation "same-state STATE1 STATE2.  STATE1 and STATE2 represent states of a <planning-problem> object.  Return true if they represent the same state."))
  

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; operations from <env>
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod sample-next ((e <planning-problem>) s a)
  (values (action-result e s a) -1))

(defmethod sample-init ((e <planning-problem>))
  (init-state e))

(defmethod is-terminal-state ((e <planning-problem>) s)
  (goal? e s))

