;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; lookahead/prop/nc-strips.lisp
;; nc-strips (nondeterministic, conditional strips) :
;; a particular representation for abstract action descriptions in 
;; propositional domains
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package lookahead)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; nondeterministic strips
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defstruct (nstrips (:constructor create-nstrips))
  "Structure type nstrips.  Represents a nondeterministic STRIPS outcome.
Create using make-nstrips with initargs
:add-list - as in strips
:delete-list - as in strips
:poss-add-list - literals that can possibly be added.
:poss-delete-list - literals that can possibly be deleted.

Each of these arguments must be a designator for a list of propositions, i.e., if it is not a list it is interpreted as a one-element list.  Note that this won't work in the case where there's a single proposition which happens to be a list, i.e., you can use 'foo, but instead of '(foo 3 4), use '((foo 3 4)).

None of the lists can intersect, except poss-add-list and poss-delete-list

An nstrips object, once created, should be treated as immutable - the lists should not be changed.  The clone method therefore just returns the original object."

  (add-list nil)
  (delete-list nil)
  (poss-add-list nil)
  (poss-delete-list nil))

(defmethod clone ((d nstrips))
  d)

(defun make-nstrips (&key (add-list nil) (delete-list nil)
			  (poss-add-list nil) (poss-delete-list nil))
  (flet ((check-valid (l)
	   (assert
	       (and (listp l) (every #'is-proposition l))
	       nil "~a is not a list of propositions" l))
	 (check-not-intersecting (l1 l2)
	   (assert
	       (notany (lambda (x) (member x l2 :test #'prop-equal)) l1)
	       nil "~a and ~a intersect" l1 l2)))
    (let ((add-list (designated-list add-list))
	  (delete-list (designated-list delete-list))
	  (poss-add-list (designated-list poss-add-list))
	  (poss-delete-list (designated-list poss-delete-list)))
      (check-valid add-list)
      (check-valid delete-list)
      (check-valid poss-add-list)
      (check-valid poss-delete-list)
      
      
      (create-nstrips :add-list add-list :delete-list delete-list
		      :poss-add-list poss-add-list 
		      :poss-delete-list poss-delete-list))))
  

(defparameter *nstrips-noop*
    (make-nstrips)
  "Object of type nstrips that represents doing nothing.  So all four lists are empty.")


(defmethod successor-set ((desc nstrips) (s <dnf-set>))
  (make-dnf-set
   (nstrips-result desc (formula s))
   (props s)))
  
(defun nstrips-result (desc f)
  (inter-clause-simplify 
   (mapcar 
    #'(lambda (c) (nstrips-clause-result desc c))
    (dnf-clauses f))))



(defun nstrips-clause-result (desc c)
  (when c
    
    ;; extract the set of literals in this clause
    (let ((literals (cond ((eql c t) nil)
			  ((is-literal c) (list c))
			  (t (rest c))))
	  (new-literals nil)
	  (adds (nstrips-add-list desc))
	  (dels (nstrips-delete-list desc))
	  (poss-adds (nstrips-poss-add-list desc))
	  (poss-dels (nstrips-poss-delete-list desc)))
      
      (dolist (l literals)
	(when (is-literal l)
	(let ((p (literal-prop l)))
	  (unless 
	      (or (member p adds :test #'prop-equal) 
		  (member p dels :test #'prop-equal)
		  (and (member p poss-adds :test #'prop-equal) (is-negation l))
		  (and (member p poss-dels :test #'prop-equal) (not (is-negation l))))
	    (push l new-literals)))))
      
      (dolist (p adds) (push p new-literals))
      (dolist (p dels) (push (negate p) new-literals))
      (conjoin-list new-literals))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; nondeterministic, conditional STRIPS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; basic defs
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defstruct (ncstrips-clause (:conc-name ncc-) (:constructor make-ncstrips-clause (precond effect)))
  precond
  effect)

(defstruct (ncstrips (:constructor create-ncstrips (clauses)))
  "Structure type ncstrips.  Create using make-ncstrips.  Once created, should be treated as immutable."
  clauses)

(defmethod clone ((d ncstrips))
  d)

(defun make-ncstrips (&rest args)
  "make-ncstrips PRECOND1 EFFECT1 ... PRECOND-N EFFECT-N 
or
make-ncstrips ((PRECOND1 . EFFECT1) ... (PRECOND-N . EFFECTN))

Each precondition must be a conjunction, and each effect is of type nstrips.  The preconditions are evaluated in order, so the first case that matches is the one whose effect happens.  At a state satisfying none of the preconditions, the action is a noop."
  (let ((arg (if (= (length args) 1) (first args) (p2alist args))))
    (let ((preconds (mapcar #'car arg))
	  (effects (mapcar #'cdr arg)))
      (assert (every #'is-dnf-clause preconds) ()
	"Preconditions ~a must all be conjunctions of literals" preconds)
      (assert (every #'nstrips-p effects) ()
	"Effects ~a must all be of type nstrips" effects)
      (create-ncstrips (mapcar #'make-ncstrips-clause preconds effects)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; computing successor states and sets
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod succ-state ((s list) (d ncstrips))
  (let ((next (copy-tree s)))
    (dolist (c (ncstrips-clauses d) next) ;; when no precondition matches, just return s
      (when (holds s (ncc-precond c))
	(let ((e (ncc-effect c)))
	  (dolist (p (nstrips-add-list e))
	    (adjoinf next p :test #'prop-equal))
	  (dolist (p (nstrips-delete-list e))
	    (deletef next p :test #'prop-equal))
	  (return-from succ-state next))))))

(defmethod succ-state ((s prop-domain-state) (d ncstrips))
  (make-prop-domain-state
   :domain (pds-domain s)
   :props (let ((l (mapset 'list #'identity (pds-props s))))
	    (dolist (c (ncstrips-clauses d))
	      (when (holds s (ncc-precond c))
		(let ((e (ncc-effect c)))
		  (dolist (p (nstrips-add-list e))
		    (adjoinf l p :test #'prop-equal))
		  (dolist (p (nstrips-delete-list e))
		    (deletef l p :test #'prop-equal))
		  (return))))
	    l)))
	    
(defmethod successor-set ((d ncstrips) (s <prop-state-set>))
  "ncstrips-result NCSTRIPS-DESCRIPTION PROP-STATE-SET.  Returns a prop-state-set representing the set of possibilities if we start at a state in the original set then do an action with this description."
  (make-prop-state-set (pss-domain s) (sequence-dnf (formula s) (ncstrips-clauses d)) :simplified t))

(defun sequence-dnf (f ncstrips-clauses &optional (simplify t))
  (let ((f
   (apply #'dnf-or
	  (mapcar #'(lambda (dnf-clause) (sequence-cnj dnf-clause ncstrips-clauses))
		  (dnf-clauses f)))))
    (if simplify (simplify-dnf f) f)))

(defun sequence-cnj (dnf-clause ncstrips-clauses)
  (when ncstrips-clauses
    (dbind (c1 . crest) ncstrips-clauses
      (let ((precond (ncc-precond c1))
	    (effect (ncc-effect c1)))
	(let ((actual-precond (simplify-dnf (dnf-and dnf-clause precond))))
	  (dnf-or (awhen actual-precond (nstrips-result effect it))
		  (sequence-dnf (dnf-and dnf-clause (dnf-not precond)) crest nil)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; regression
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defmethod regress (s1 s2 desc)
  (make-prop-state-set (pss-domain s1) (regress-dnf-ncstrips (formula s1) (formula s2) desc)))

(defun regress-dnf-ncstrips (f1 f2 desc)
  "regress-dnf-ncstrips FORMULA1 FORMULA2 NCSTRIPS-DESC
Return a clause that implies FORMULA1, such that progressing any state compatible with the clause yields a formula compatible with FORMULA2."
  
  (dolist (clause (dnf-clauses f1) nil)
    (awhen (regress-cnj-ncstrips clause f2 (etypecase desc (ncstrips (ncstrips-clauses desc)) (list desc)))
	   (return (simplify-dnf it)))))

(defun regress-cnj-ncstrips (clause f2 clauses)
  (dolist (nclause clauses nil)
    (let ((precond (ncc-precond nclause))
	  (effect (ncc-effect nclause)))
      (awhen (regress-dnf-nstrips effect (dnf-and clause precond) f2)
	     (return it)))))

(defun regress-dnf-nstrips (desc clause formula)
  (dolist (next-clause (dnf-clauses formula) nil)
    (awhen (regress-clause-nstrips desc clause next-clause)
	   (return it))))


(defun regress-clause-nstrips (desc clause1 clause2)
  "regress-clause-nstrips NSTRIPS-DESC CLAUSE1 CLAUSE2.

Return a new clause that implies CLAUSE1, such that for any state compatible with the new clause, progressing that state through DESC yields a clause that's compatible with CLAUSE2."

  ;; step through clause2, and l3 = progress(clause1)
  (let ((conjuncts nil))
    (dolist (lit (conjuncts clause2))
      (if (is-negation lit)
	  (cond ((or (member (negatee lit) (nstrips-delete-list desc) :test #'prop-equal)
		     (member (negatee lit) (nstrips-poss-delete-list desc) :test #'prop-equal)))
		((member (negatee lit) (nstrips-add-list desc) :test #'prop-equal)
		 (return-from regress-clause-nstrips nil))
		(t (push lit conjuncts)))
	(cond ((or (member lit (nstrips-add-list desc) :test #'prop-equal)
		   (member lit (nstrips-poss-add-list desc) :test #'prop-equal)))
	      ((member lit (nstrips-delete-list desc) :test #'prop-equal)
	       (return-from regress-clause-nstrips nil))
	      (t (push lit conjuncts)))))
    (conjoin-list (concatenate 'list conjuncts (conjuncts clause1)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; schematized ncstrips descriptions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defstruct (ncstrips-schema (:conc-name nil) (:constructor create-ncstrips-schema))
  domains
  var-names
  desc
  dom
  arg-lists)


(defun make-ncstrips-schema (&key planning-problem var-domains effects)
  "make-ncstrips-schema &key PLANNING-PROBLEM VAR-DOMAINS EFFECTS

VAR-DOMAINS is an association list from symbols to sets
EFFECTS is an association list mapping from preconditions, which are formulae, to nstrips descriptions."
  (create-ncstrips-schema :domains var-domains
			  :dom planning-problem
			  :desc (make-ncstrips effects)
			  :var-names (mapcar #'car var-domains)
			  :arg-lists (apply #'direct-product 'list (mapcar #'cdr var-domains))))


(defmacro make-ncstrips-schemas (dom &rest args)
  "make-ncstrips-schemas PLANNING-PROB &rest SPECS

Definition macro for creating a set of ncstrips schemas.
PLANNING-PROB is a planning problem of type <prop-domain>
Each SPEC_i is a list of the form (NAME :var-domains VAR-DOMAINS :effects EFFECTS) where
- NAME is a symbol
- VAR-DOMAINS looks like ((VAR1 DOM1) ... (VAR_n DOM_n)) where VAR_i is a symbol, and DOM_i is a symbol naming a set (according to the mapping in PLANNING-PROB)
- EFFECTS is a list of effects.

Each effect in EFFECTS is a key-value list with some of the following keys
- PRECOND is a literal or dnf clause with free variables among the VAR_i above.  Defaults to t.
- Each of ADD-LIST, DELETE-LIST, POSS-ADD-LIST, POSS-DELETE-LIST is a list, where each element is either 1) a positive literal 2) a list of the form (for-all ((VAR_1 DOM_1) ... (VAR_k DOM_k)) FORMULA LIT), where the VAR_k and DOM_k are interpreted as above, FORMULA is a formula, and LIT is a positive literal having free variables among the new variables and the original ones.  Each of the four lists defaults to nil.
  
The meaning of the second form for the effect lists is that for each set of variable bindings in the given sets that satisfy the formula, we will add a corresponding occurrence of LIT to the given list.  Note that this is done independent of the state, i.e., formula must only refer to literals that are nonfluents for the given planning problem.

Another addition is that in the precondition, in any position where a term would normally be, you can put in a list, which is then treated as a form and evaluated in the lexical context where d is bound.  This allows referring to constant parameters of the domain."
  `(list
    ,@(mapcar
       #'(lambda (spec)
	   `(cons ',(car spec)
		  ,(lookup-types-in-schema-def (cdr spec) dom)))
       args)))

(defun lookup-types-in-schema-def (def dom)
  (labels ((replace-type (x)
	     `(lookup-type ,dom ',x))
	   (replace-terms (x)
	     `(list ',(prop-symbol x)
		    ,@(mapcar
		       #'(lambda (term)
			   (etypecase term
			     (symbol `',term)
			     (list term)))
		       (prop-args x))))
	       
	   (replace-effect-types (l)
	     "replace symbols in for-alls"
	     `(list
	     ,@(mapcar #'(lambda (x)
			   (if (and (listp x) (eq (first x) 'for-all))
			       (dbind (lists formula lit) (rest x)
				 `(list 'for-all
					(list
					 ,@(mapcar #'(lambda (l)
						     `(cons ',(first l)
							    (lookup-type ,dom ',(second l))))
						   lists))
					',formula
					',lit))
			   `',x))
		       l)))
	   (replace-precond-types (precond)
	     "replace for-alls, function calls in precondition"
	     (etypecase precond
	       ((member t nil) precond)
	       ((satisfies is-literal) (replace-precond-types (conjoin precond)))
	       ((satisfies is-conjunction)
		(let ((v (gensym)))
		`(conjoin-list
		  (append
		   ,@(mapcar #'(lambda (x)
			       (if (and (listp x) (eq (car x) 'for-all))
				   (dbind ((var vals) prop) (cdr x)
				     `(mapset 'list #'(lambda (,v) (bind ,(replace-terms prop) (list (cons ',var ,v))))
					      (lookup-type ,dom ',vals)))
				 `(list ,(replace-terms x))))
			   (conjuncts precond)))))))))
		 
	     
			     
    `(make-ncstrips-schema
	 :planning-problem ,dom
	 :var-domains (list
		       ,@(mapcar #'(lambda (pair)
				     `(cons ',(first pair) ,(replace-type (second pair))))
				 (second def)))
	 :effects (list
		   ,@(mapcar #'(lambda (eff)
				 (dbind (&key (precond nil) (poss-add-list nil) (add-list nil)
					      (delete-list nil) (poss-delete-list nil)) eff
				   `(cons 
				     ,(replace-precond-types precond)
				     (make-nstrips
				      :add-list ,(replace-effect-types add-list)
				      :delete-list ,(replace-effect-types delete-list)
				      :poss-add-list ,(replace-effect-types poss-add-list)
				      :poss-delete-list ,(replace-effect-types poss-delete-list)))))
			     (fourth def))))))
       
  
  
  




(defun instantiate (schema arg-list)
  "instantiate SCHEMA ARG-LIST
SCHEMA is an ncstrips schema
ARG-LIST is the list of arguments."
  ;; TODO this is quite inefficiently done.  Instead, should cache how to do this
  ;; when the schema is being created.
  (let ((bindings (mapcar #'cons (var-names schema) arg-list))
	(dom (dom schema)))
    (flet ((set-vars (f)
	     (if (and (listp f) (eq (first f) 'for-all))
		 (dbind (lists condition lit) (rest f)
		   (let ((names (mapcar #'car lists))
			 (doms (mapcar #'cdr lists)))
		     (mapset 'list
			     #'(lambda (vals)
				 (bind lit (append (mapcar #'cons names vals) bindings)))
			     (filter 'list (apply #'direct-product 'list doms)
				     #'(lambda (vals)
					 (holds-background 
					  dom (bind condition (append (mapcar #'cons names vals) bindings))))))))
				
	       (list (bind f bindings)))))
      (make-ncstrips 
       (mapcan
	#'(lambda (clause)
	    (let ((pre (bind (ncc-precond clause) bindings)))
	      (check-type pre (satisfies is-conjunction))
	      (setf pre
		(let ((conjuncts
			(mapcar #'(lambda (lit)
				    (if (eq (prop-type (dom schema) lit) 'nonfluent)
					(holds-background dom lit)
				      lit))
				(conjuncts pre))))
		  (unless (member nil conjuncts) (conjoin-list conjuncts))))
	      (when pre
		(list
		 (cons pre
		       (let ((e (ncc-effect clause)))
			 (make-nstrips 
			  :add-list (mapcan #'set-vars (nstrips-add-list e))
			  :delete-list (mapcan #'set-vars (nstrips-delete-list e))
			  :poss-add-list (mapcan #'set-vars (nstrips-poss-add-list e))
			  :poss-delete-list (mapcan #'set-vars (nstrips-poss-delete-list e)))))))))
	(ncstrips-clauses (desc schema)))))))

    



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Standard STRIPS (implemented in terms of NCSTRIPS)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun make-strips (&key precond add-list delete-list)
  "make-strips &key PRECOND ADD-LIST DELETE-LIST
PRECOND - a conjunction of positive literals
ADD-LIST, DELETE-LIST - lists of positive literals."
  (make-ncstrips (p2alist precond (make-nstrips :add-list add-list :delete-list delete-list))))

(defun is-strips (desc)
  "is-strips NCSTRIPS-DESC.  Returns t if the description has a single deterministic clause whose precondition is a conjunction of positive literals."
  (check-type desc ncstrips)
  (let* ((clauses (ncstrips-clauses desc)))
    (unless (or (null clauses) (length-exceeds clauses 1))
      (let ((eff (ncc-effect (first clauses)))
	    (precond (ncc-precond (first clauses))))
	(and
	 (or (is-conjunction precond) (is-literal precond))
	 (every #'is-proposition (conjuncts precond))
	 (null (nstrips-poss-add-list eff))
	 (null (nstrips-poss-delete-list eff)))))))

(deftype strips ()
  "subtype of ncstrips that satisfies is-strips"
  `(satisfies is-strips))

(defun precond (desc)
  "precond DESC.  DESC is required to be a STRIPS description."
  (check-type desc strips)
  (ncc-precond (first (ncstrips-clauses desc))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; schemas
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun make-strips-schema (&key domain domains precond add-list delete-list)
  "make-strips-schema &key DOMAINS PRECOND ADD-LIST DELETE-LIST. 
DOMAIN is of type <prop-domain>
DOMAINS are as in make-ncstrips-schema
PRECOND is a conjunction of positive literals
ADD-LIST and DELETE-LIST are lists of positive literals."
  (make-ncstrips-schema :var-domains domains :planning-problem domain
			:effects (p2alist precond (make-nstrips :add-list add-list :delete-list delete-list))))



