;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; rl/rl-user.lisp
;; contains functions to make it easier for users to call the reinforcement
;; learning code
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package rl)


(defun learn (env policy obs num-steps &key (hist-length nil) (step-print-inc 100) (ep-print-inc 10) (reset t))
  "learn ENV POLICY OBSERVERS NUM-STEPS &key (HIST-LENGTH nil) (STEP-PRINT-INC 100) (EP-PRINT-INC 10) (RESET t)

A frontend for running trajectory-based rl as defined in the <rl-control> class.  
ENV - type <env>
POLICY - Either an object of type [policy], or the symbol 'rl:random, which means choose uniformly at random from available actions.
OBSERVERS - either an <rl-observer> or a list of <rl-observers>.  <rl-observers> include learning algorithms as well as objects that gather statistics, etc.  
NUM-STEPS - Number of env steps to go for.

Keys
HIST-LENGTH - Length of history to gather.  Nil by default, which means don't gather history.
STEP-PRINT-INC - Print acknowledgement every so many env steps
EP-PRINT-INC - Print acknowledgement every so many episodes
RESET - Whether to reset all the learning algorithms beforehand."
  
  (when (eq policy 'random)
    (setf policy (policy:make-random-policy (lambda (s) (env:avail-actions env s)))))
  (unless (listp obs)
    (setf obs (list obs)))
  (dolist (o obs)
    (when (typep o '<learning-algorithm>)
      (when reset
	(reset o))
      (set-hist-collect 
       (make-evenly-spaced-intervals 1 num-steps hist-length)
       o)
      (when (typep o '<on-policy-learning-algorithm>)
	(assert (eq o policy) nil
	  "Observer ~a is an on-policy learning algorithm, but it does not equal the exploration policy ~a" o policy))))
    
  (push (progress-printer:make-progress-printer step-print-inc ep-print-inc)
	obs)
  (reinforcement-learning (make-instance '<rl-control>
				:env env
				:policy policy
				:obs obs
				:num-steps num-steps)))


(defun on-policy-learning (env alg &rest args)
  "on-policy-learning ENV ON-POLICY-ALG &rest ARGS.  Like calling (learn ENV ALG ALG &rest ARGS).  Strictly speaking, this is not exactly on-policy learning.  Rather, it just means that the learning algorithm is of type <on-policy-learning-algorithm>, and may rely on the fact that its make-choice method will be called right before each environment step to choose the next action."
  (let ((*on-policy-learning* alg))
    (apply #'learn env alg alg args)))

  
  
  

(defun evaluate (env policies &key (num-steps nil) (step-print-inc 100) (env-display-stream nil) (pause-after-actions nil)
				   (num-trials 1))
  "evaluate ENV POLICIES &key (NUM-TRIALS 1) (NUM-STEPS nil) (STEP-PRINT-INC 100) (ENV-DISPLAY-STREAM nil) (PAUSE-AFTER-ACTIONS nil)

Front end for evaluating policies.  POLICIES is either a policy or a sequence of policies.  Do the following for each policy : reset environment to a random state, and run the given policy, computing the total reward, until either the episode terminates or NUM-STEPS environment steps happen.  Repeat this NUM-TRIALS times.   If POLICIES was a sequence, return a corresponding sequence of average total rewards rounded to 2 decimal places.  If POLICIES was a singleton, just return a number. Acknowledgement is printed every STEP-PRINT-INC steps, if it is non-nil.  If ENV-DISPLAY-STREAM is non-nil, then print what happens in environment to this stream.  If, in addition, PAUSE-AFTER-ACTIONS is non-nil, then ask user to press enter after each step.  Future todo : make it so a set of initial states is sampled, and those initial states are used to evaluate each policy."
  
  (let* ((pp (progress-printer:make-progress-printer step-print-inc nil))
	 (sg (make-instance 'stat-gatherer:<stat-gatherer>))
	 (obs (list sg pp)))

    (awhen env-display-stream
	   (push (env-observer:make-env-observer it pause-after-actions) obs))
    
    (when step-print-inc
      (format t "~&Evaluating policies"))
    (flet ((evaluate-policy (pol)
	     (let ((rlc (make-instance '<rl-control>
			  :env env
			  :policy pol
			  :obs obs
			  :num-steps num-steps
			  :num-episodes 1)))
	       (format t ".")
	       (stat-gatherer:reset-counts sg)
	       (dotimes (j num-trials)
		 (reinforcement-learning rlc)
		 
		 )
	       (round-decimal
		(/ (stat-gatherer:total-reward sg) num-trials)
		2))))


      (typecase policies
	(set:[numbered-set] (set:mapset 'vector #'evaluate-policy policies))
	(otherwise (evaluate-policy policies))))))
      

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; gathering trajectories
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun gather-trajectories (env pol &key (num-episodes nil))
  "gather-trajectories ENV POLICY &key (NUM-EPISODES nil)

Return the trajectories resulting from running NUM-EPISODES using POLICY.  See trajectory-gatherer:trajectories for format."
  (let ((tg (make-instance 'trajectory-gatherer:<trajectory-gatherer>)))
    (reinforcement-learning (make-instance '<rl-control> :env env :policy pol :num-episodes num-episodes :obs (list tg)))
    (trajectory-gatherer:trajectories tg)))

(defun state-trajectory (trajectory &key (result-type 'vector))
  "state-trajectory TRAJECTORY.  Given a trajectory S1 A1 R1 S2 ... Sn, return the state trajectory alone."
  (extract-subsequence trajectory 0 3 :result-type result-type))

(defun action-trajectory (trajectory &key (result-type 'vector))
  "state-trajectory TRAJECTORY.  Given a trajectory S1 A1 R1 S2 ... Sn, return the action trajectory alone."
  (extract-subsequence trajectory 1 3 :result-type result-type))

(defun transitions (trajectory &key (share-structure t))
  "Implicit representation of the set of transitions in this trajectory, each represented as a list (s a r s2).
The lists share structure by default."
  (assert share-structure)
  (let ((l (make-list 4)))
    (ndlet ((i (/ (1- (length trajectory)) 3)))
      (let ((k (* i 3)))
	(setf (first l) (aref trajectory k)
	      (second l) (aref trajectory (1+ k))
	      (third l) (aref trajectory (+ 2 k))
	      (fourth l) (aref trajectory (+ 3 k)))
	l))))

(defun all-transitions (trajectories &key (share-structure t))
  (assert share-structure)
  (ndunion ((tr trajectories))
    (transitions tr :share-structure share-structure)))
  

(defmacro do-transitions (((s a r s2) traj) &rest body)
  "do-transitions ((S A R S2) TRAJ) &rest BODY

S,A,R,S2 (not evaluated): symbols that will be bound to state, action, reward, and successor state in transitions in this trajectory.
TRAJ (evaluated): a vector of the form S,A,R,S2,A2,R2,...,Sn
BODY (unevaluated): list of forms"
  
  ;; could be simplified given transitions above
  (with-gensyms (l tr i k)
    `(let* ((,tr ,traj)
	    (,l (length ,tr)))
       (dotimes (,i (/ (1- ,l) 3))
	 (let ((,k (* ,i 3)))
	   (let ((,s (aref ,tr ,k))
		 (,a (aref ,tr (1+ ,k)))
		 (,r (aref ,tr (+ 2 ,k)))
		 (,s2 (aref ,tr (+ 3 ,k))))
	     (declare (ignorable ,s ,a ,r ,s2))
	     ,@body))))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; io interface
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun io-interface (env &optional (advisors nil) (algs nil))
  "io-interface ENV &optional (ADVISORS nil) (ALGS nil)

ENV: environment
ADVISORS: designator for a list of advisors
ALGS: designator for a list of observers (usually learning algorithms)

Step through the environment in an interactive mode where actions are chosen by asking the user, and events are printed to the terminal.  Each advisor in ADVISORS can be a policy, Q-function, or any object that implements the print-advice method, and is asked to print advice after each step.  For any of the observers in ALGS that are also learning algorithms, run them and set their debug stream to *standard-output* beforehand.  State history is stored in *hist*."
  
  (setf *hist* (make-array 0 :adjustable t :fill-pointer 0))
  (let ((pol 
	 (make-instance 'policy:<prompt-policy>
	   :choice-fn (lambda (s) (env:avail-actions env s))
	   :advisors (designated-list advisors)
	   :io-specifier env))
	(alg-list (designated-list algs)))
    (let ((old-vals (mapcar
		     #'(lambda (obs) (when (typep obs '<learning-algorithm>) (debug-str obs)))
		     alg-list)))
      (dolist (alg alg-list)
	(when (typep alg '<learning-algorithm>)
	  (rl-user:set-debug-str *standard-output* alg)))
      (push (make-instance 'env-observer:<env-observer> :prompt-after-episodes t) alg-list)
      (reinforcement-learning (make-instance '<rl-control>
				:env env
				:policy pol
				:obs alg-list))
      (mapcar #'(lambda (alg old-val) (when old-val (rl-user:set-debug-str old-val alg)))
	      (rest alg-list) old-vals)
      (values))))


(defmethod env-user:io-interface ((env t))
  (io-interface env))

(defun batch-learning (obs &key (hist-length nil) (step-print-inc 100) (ep-print-inc 10) (reset t) (episode-lengths nil) (query-stream nil)
				(filename nil) (input-stream nil))
  "batch-learning OBSERVERS &key (HIST-LENGTH nil) (STEP-PRINT-INC 100) (EP-PRINT-INC 10) (RESET t) (EPISODE-LENGTHS nil) (FILENAME nil) (QUERY-STREAM nil)
OBSERVERS - a designator for a list of <rl-observer> objects.
HIST-LENGTH - a positive integer or nil.  Represents the number of times each learning algorithm collects a snapshot of its state.  nil means 0.
STEP-PRINT-INC, EP-PRINT-INC - positive integers.  How often to print acknowledgement to the terminal.
RESET - boolean.  Whether to reset all the observers before beginning.

EPISODE-LENGTHS - list of positive integers.  Number of state transitions seen in each episode.
FILENAME - file that contains the episodes, separated by newlines.  Each line looks like S1 A1 R1 ... Sn.  
INPUT-STREAM - if non-nil, this is the stream from which trajectories are read.
QUERY-STREAM - if non-nil, a prompt message will be sent to this stream before reading each object from STREAM.
Exactly one of the EPISODE-LENGTHS and FILENAME arguments must be provided.  If FILENAME is provided, INPUT-STREAM and QUERY-STREAM must not be provided.

"
  ;; Note this doesn't use or subclass <rl-control>, suggesting that that abstraction isn't quite right
  (assert (xor episode-lengths filename) () "Exactly one of the episode-lengths and filename arguments must be given.")
  (assert (not (and filename (or query-stream input-stream))) () "Query stream or input stream cannot be given because input is being read from file.")
  
  
  ;; If necessary figure out episode lengths from file
  (orf episode-lengths
       (with-open-file (str filename :type :input)
	 (loop
	     for next-line = (read-line str nil 'eof)
	     until (eq next-line 'eof)
	     collect (with-input-from-string (s next-line)
		       (let ((i 0))
			 (loop
			   (let ((x (read s nil 'eof)))
			     (when (eq x 'eof)
			       (return (/ (1- i) 3)))
			     (incf i))))))))

  ;; Set up observers
  (let ((obs (designated-list obs))
	(num-steps (reduce #'+ episode-lengths)))
    (dolist (o obs)
      (when (typep o '<learning-algorithm>)
	(when reset
	  (reset o))
	(set-hist-collect
	 (make-evenly-spaced-intervals 1 num-steps hist-length)
	 o)
	(assert (not (typep o '<on-policy-learning-algorithm>)) nil "On-policy algorithm ~a not allowed for batch learning" o)))
    (push (progress-printer:make-progress-printer step-print-inc ep-print-inc) obs)

    ;; Helper function to actually read the episodes from a stream
    (flet ((read-episodes-from-stream (str)
	     (macrolet ((broadcast (msg &rest args)
			  (with-gensyms (o)
			    `(dolist (,o obs)
			       (,msg ,o ,@args))))
			(prompt (msg &rest args)
			  `(awhen query-stream 
				  (fresh-line it)
				  (format it ,msg ,@args))))
      
	       ;; Begin
	       (broadcast inform-start-execution)
    
	       ;; Loop over episodes
	       (dolist (l episode-lengths)
      
		 ;; Begin episode
		 (prompt "Enter initial state. ")
		 (let ((s (read str)))
		   (broadcast inform-start-episode s))
      
		 ;; Loop over transitions
		 (dotimes (i l)
		   (let (a r s2)
	  
		     ;; read in this transition from the stream

		     (prompt "Enter action. ")
		     (setf a (read str))
		     (prompt "Enter reward. ")
		     (setf r (read str))
		     (prompt "Enter next state. ")
		     (setf s2 (read str))
	  
		     ;; inform observers about this transition, and termination if necessary
		     (broadcast inform-env-step a r s2 (if (eq i (1- l)) t nil))
		     )))
    
	       ;; Finish
	       (broadcast inform-finish-execution))))

      ;; Main body
      (if filename
	  (with-open-file (str filename :type :input)
	    (read-episodes-from-stream str))
	(read-episodes-from-stream (check-not-null input-stream))))))
  


      
      
	  
  
  
